import { StdBatchFactory } from './StdBatchFactory';
import { AggregateUniformsBatch } from './AggregateUniformsBatch';
import * as PIXI from 'pixi.js';
export class AggregateUniformsBatchFactory extends StdBatchFactory {
    constructor(renderer) {
        super(renderer);
        this.MAX_UNIFORMS = Math.floor(Math.min(renderer.renderer.gl.getParameter(renderer.renderer.gl.MAX_VERTEX_UNIFORM_VECTORS), renderer.renderer.gl.getParameter(renderer.renderer.gl.MAX_FRAGMENT_UNIFORM_VECTORS))
            / (4 * renderer._uniformRedirects.length));
        this.uniformBuffer = this._createUniformBuffer();
        this.uniformMap = [];
        this.uniformLength = 0;
    }
    _newBatch() {
        const batch = new AggregateUniformsBatch(this._renderer);
        batch.uniformBuffer = this._createUniformBuffer();
        batch.uniformMap = [];
        return batch;
    }
    _put(displayObject) {
        if (!this._renderer._uniformIDAttrib) {
            if (this.uniformLength >= 0) {
                const id = this._matchUniforms(displayObject);
                if (id >= 0) {
                    return true;
                }
                return false;
            }
        }
        if (this.uniformLength + 1 >= this.MAX_UNIFORMS) {
            return false;
        }
        if (this._renderer._uniformIDAttrib) {
            const id = this._matchUniforms(displayObject);
            if (id >= 0) {
                this.uniformMap.push(id);
                return true;
            }
        }
        for (let i = 0, j = this._renderer._uniformRedirects.length; i < j; i++) {
            const uniformRedirect = this._renderer._uniformRedirects[i];
            const { source, glslIdentifer } = uniformRedirect;
            const value = typeof source === 'string'
                ? displayObject[source]
                : source(displayObject, this._renderer);
            if (Array.isArray(value)) {
                this.uniformBuffer[glslIdentifer].push(...value);
            }
            else {
                this.uniformBuffer[glslIdentifer].push(value);
            }
        }
        this.uniformMap.push(this.uniformLength);
        ++this.uniformLength;
        return true;
    }
    _buildBatch(batch) {
        super._buildBatch(batch);
        const buffer = batch.uniformBuffer;
        const map = batch.uniformMap;
        batch.uniformBuffer = this.uniformBuffer;
        batch.uniformMap = this.uniformMap;
        batch.uniformLength = this.uniformLength;
        this.uniformBuffer = buffer;
        this.uniformMap = map;
        this.uniformLength = 0;
        this._resetUniformBuffer(this.uniformBuffer);
        this.uniformMap.length = 0;
    }
    _createUniformBuffer() {
        const buffer = {};
        for (let i = 0, j = this._renderer._uniformRedirects.length; i < j; i++) {
            const uniformRedirect = this._renderer._uniformRedirects[i];
            buffer[uniformRedirect.glslIdentifer] = [];
        }
        return buffer;
    }
    _resetUniformBuffer(buffer) {
        for (let i = 0, j = this._renderer._uniformRedirects.length; i < j; i++) {
            const uniformRedirect = this._renderer._uniformRedirects[i];
            buffer[uniformRedirect.glslIdentifer].length = 0;
        }
    }
    _matchUniforms(displayObject) {
        const uniforms = this._renderer._uniformRedirects;
        for (let i = this.uniformLength - 1; i >= 0; i--) {
            let isMatch = true;
            for (let k = 0, n = uniforms.length; k < n; k++) {
                const { glslIdentifer, source } = uniforms[k];
                const value = typeof source === 'string'
                    ? displayObject[source]
                    : source(displayObject, this._renderer);
                if (!this._compareUniforms(value, this.uniformBuffer[glslIdentifer][i])) {
                    isMatch = false;
                    break;
                }
            }
            if (isMatch) {
                return i;
            }
        }
        return -1;
    }
    _compareUniforms(u1, u2) {
        if (u1 === u2) {
            return true;
        }
        if (u1.group || u2.group) {
            return false;
        }
        if (u1.equals) {
            return u1.equals(u2);
        }
        if (Array.isArray(u1) && Array.isArray(u2)) {
            if (u1.length !== u2.length) {
                return false;
            }
            for (let i = 0, j = u1.length; i < j; i++) {
                if (u1[i] !== u2[i]) {
                    return false;
                }
            }
            return true;
        }
        if (u1 instanceof PIXI.Point && u2 instanceof PIXI.Point) {
            return u1.x === u2.x && u1.y === u2.y;
        }
        if (u1 instanceof PIXI.Matrix && u2 instanceof PIXI.Matrix) {
            return u1.a === u2.a && u1.b === u2.b
                && u1.c === u2.c && u1.d === u2.d
                && u1.tx === u2.tx && u1.ty === u2.ty;
        }
        if (u1 instanceof PIXI.BaseTexture && u2 instanceof PIXI.BaseTexture) {
            return u1.uid === u2.uid;
        }
        return false;
    }
}
//# sourceMappingURL=AggregateUniformsBatchFactory.js.map