import { AttributeRedirect } from './redirects/AttributeRedirect';
import * as PIXI from 'pixi.js';
export class BatchGeometry extends PIXI.Geometry {
    constructor(attributeRedirects, hasIndex, texIDAttrib, texturesPerObject, inBatchIDAttrib, uniformIDAttrib, masterIDAttrib, attributeBuffer, indexBuffer) {
        super();
        attributeBuffer = attributeBuffer || new PIXI.Buffer(null, false, false);
        indexBuffer = indexBuffer || (hasIndex ? new PIXI.Buffer(null, false, true) : null);
        attributeRedirects.forEach((redirect) => {
            const { glslIdentifer, glType, glSize, normalize } = redirect;
            this.addAttribute(glslIdentifer, attributeBuffer, glSize, normalize, glType);
        });
        if (!masterIDAttrib) {
            if (texIDAttrib && texturesPerObject > 0) {
                this.addAttribute(texIDAttrib, attributeBuffer, texturesPerObject, true, PIXI.TYPES.FLOAT);
            }
            if (inBatchIDAttrib) {
                this.addAttribute(inBatchIDAttrib, attributeBuffer, 1, false, PIXI.TYPES.FLOAT);
            }
            if (uniformIDAttrib) {
                this.addAttribute(uniformIDAttrib, attributeBuffer, 1, false, PIXI.TYPES.FLOAT);
            }
        }
        else {
            this.addAttribute(masterIDAttrib, attributeBuffer, 1, false, PIXI.TYPES.FLOAT);
        }
        if (hasIndex) {
            this.addIndex(indexBuffer);
        }
        this.attribBuffer = attributeBuffer;
        this.indexBuffer = indexBuffer;
    }
}
export class IBatchGeometryFactory {
    constructor(renderer) {
        this._renderer = renderer;
    }
}
export class BatchGeometryFactory extends IBatchGeometryFactory {
    constructor(renderer) {
        super(renderer);
        this._targetCompositeAttributeBuffer = null;
        this._targetCompositeIndexBuffer = null;
        this._aIndex = 0;
        this._iIndex = 0;
        this._attribRedirects = renderer._attribRedirects;
        this._indexProperty = renderer._indexProperty;
        this._vertexCountProperty = renderer._vertexCountProperty;
        this._vertexSize = AttributeRedirect.vertexSizeFor(this._attribRedirects);
        this._texturesPerObject = renderer._texturesPerObject;
        this._textureProperty = renderer._textureProperty;
        this._texIDAttrib = renderer._texIDAttrib;
        this._inBatchIDAttrib = renderer._inBatchIDAttrib;
        this._uniformIDAttrib = renderer._uniformIDAttrib;
        this._masterIDAttrib = renderer._masterIDAttrib;
        if (!this._masterIDAttrib) {
            this._vertexSize += this._texturesPerObject * 4;
            if (this._inBatchIDAttrib) {
                this._vertexSize += 4;
            }
            if (this._uniformIDAttrib) {
                this._vertexSize += 4;
            }
        }
        else {
            this._vertexSize += 4;
        }
        if (this._texturesPerObject === 1) {
            this._texID = 0;
        }
        else if (this._texturesPerObject > 1) {
            this._texID = new Array(this._texturesPerObject);
        }
        this._aBuffers = [];
        this._iBuffers = [];
        this._geometryPool = [];
    }
    init(verticesBatched, indiciesBatched) {
        this._targetCompositeAttributeBuffer = this.getAttributeBuffer(verticesBatched);
        if (this._indexProperty) {
            this._targetCompositeIndexBuffer = this.getIndexBuffer(indiciesBatched);
        }
        this._aIndex = this._iIndex = 0;
    }
    append(targetObject, batch_) {
        const batch = batch_;
        const tex = targetObject[this._textureProperty];
        if (this._texturesPerObject === 1) {
            const texUID = tex.baseTexture ? tex.baseTexture.uid : tex.uid;
            this._texID = batch.uidMap[texUID];
        }
        else if (this._texturesPerObject > 1) {
            let _tex;
            for (let k = 0; k < tex.length; k++) {
                _tex = tex[k];
                const texUID = _tex.BaseTexture ? _tex.baseTexture.uid : _tex.uid;
                this._texID[k] = batch.uidMap[texUID];
            }
        }
        if (this._inBatchIDAttrib || this._uniformIDAttrib) {
            this._inBatchID = batch.batchBuffer.indexOf(targetObject);
        }
        if (this._uniformIDAttrib) {
            this._uniformID = batch.uniformMap[this._inBatchID];
        }
        this.geometryMerger(targetObject, this);
    }
    build() {
        const geom = (this._geometryPool.pop() || new BatchGeometry(this._attribRedirects, true, this._texIDAttrib, this._texturesPerObject, this._inBatchIDAttrib, this._uniformIDAttrib, this._masterIDAttrib));
        geom.attribBuffer.update(this._targetCompositeAttributeBuffer.float32View);
        geom.indexBuffer.update(this._targetCompositeIndexBuffer);
        return geom;
    }
    release(geom) {
        this._geometryPool.push(geom);
    }
    get geometryMerger() {
        if (!this._geometryMerger) {
            this._geometryMerger = new GeometryMergerFactory(this).compile();
        }
        return this._geometryMerger;
    }
    set geometryMerger(func) {
        this._geometryMerger = func;
    }
    getAttributeBuffer(size) {
        const roundedP2 = PIXI.utils.nextPow2(size);
        const roundedSizeIndex = PIXI.utils.log2(roundedP2);
        const roundedSize = roundedP2;
        if (this._aBuffers.length <= roundedSizeIndex) {
            this._aBuffers.length = roundedSizeIndex + 1;
        }
        let buffer = this._aBuffers[roundedSizeIndex];
        if (!buffer) {
            this._aBuffers[roundedSizeIndex] = buffer = new PIXI.ViewableBuffer(roundedSize * this._vertexSize);
        }
        return buffer;
    }
    getIndexBuffer(size) {
        const roundedP2 = PIXI.utils.nextPow2(Math.ceil(size / 12));
        const roundedSizeIndex = PIXI.utils.log2(roundedP2);
        const roundedSize = roundedP2 * 12;
        if (this._iBuffers.length <= roundedSizeIndex) {
            this._iBuffers.length = roundedSizeIndex + 1;
        }
        let buffer = this._iBuffers[roundedSizeIndex];
        if (!buffer) {
            this._iBuffers[roundedSizeIndex] = buffer = new Uint16Array(roundedSize);
        }
        return buffer;
    }
}
const CompilerConstants = {
    INDICES_OFFSET: '__offset_indices_',
    FUNC_SOURCE_BUFFER: 'getSourceBuffer',
    packerArguments: [
        'targetObject',
        'factory',
    ],
};
const GeometryMergerFactory = class {
    constructor(packer) {
        this.packer = packer;
    }
    compile() {
        const packer = this.packer;
        let packerBody = `
            const compositeAttributes = factory._targetCompositeAttributeBuffer;
            const compositeIndices = factory._targetCompositeIndexBuffer;
            let aIndex = factory._aIndex;
            let iIndex = factory._iIndex;
            const textureId = factory._texID;
            const attributeRedirects = factory._attribRedirects;
        `;
        packer._attribRedirects.forEach((redirect, i) => {
            packerBody += `
                let __offset_${i} = 0;
                const __buffer_${i} = (
                    ${this._compileSourceBufferExpression(redirect, i)});
            `;
        });
        packerBody += `
            const {
                int8View,
                uint8View,
                int16View,
                uint16View,
                int32View,
                uint32View,
                float32View,
            } = compositeAttributes;

            const vertexCount = ${this._compileVertexCountExpression()};

            let adjustedAIndex = 0;

            for (let vertexIndex = 0; vertexIndex < vertexCount; vertexIndex++)
            {
        `;
        let skipByteIndexConversion = false;
        for (let i = 0; i < packer._attribRedirects.length; i++) {
            const redirect = packer._attribRedirects[i];
            if (!skipByteIndexConversion) {
                packerBody += `
        adjustedAIndex = aIndex / ${this._sizeOf(i)};
                `;
            }
            if (typeof redirect.size === 'number') {
                for (let j = 0; j < redirect.size; j++) {
                    packerBody += `
        ${redirect.type}View[adjustedAIndex++] = __buffer_${i}[__offset_${i}++];
                    `;
                }
            }
            else {
                packerBody += `
        ${redirect.type}View[adjustedAIndex++] = __buffer_${i};
                `;
            }
            if (packer._attribRedirects[i + 1] && (this._sizeOf(i + 1) !== this._sizeOf(i))) {
                packerBody += `
        aIndex = adjustedAIndex * ${this._sizeOf(i)};
                `;
            }
            else {
                skipByteIndexConversion = true;
            }
        }
        if (skipByteIndexConversion) {
            if (this._sizeOf(packer._attribRedirects.length - 1) !== 4) {
                packerBody += `
        aIndex = adjustedAIndex * ${this._sizeOf(packer._attribRedirects.length - 1)}
                `;
                skipByteIndexConversion = false;
            }
        }
        if (!packer._masterIDAttrib) {
            if (packer._texturesPerObject > 0) {
                if (packer._texturesPerObject > 1) {
                    if (!skipByteIndexConversion) {
                        packerBody += `
            adjustedAIndex = aIndex / 4;
                        `;
                    }
                    for (let k = 0; k < packer._texturesPerObject; k++) {
                        packerBody += `
            float32View[adjustedAIndex++] = textureId[${k}];
                        `;
                    }
                    packerBody += `
            aIndex = adjustedAIndex * 4;
                    `;
                }
                else if (!skipByteIndexConversion) {
                    packerBody += `
            float32View[aIndex / 4] = textureId;
                    `;
                }
                else {
                    packerBody += `
            float32View[adjustedAIndex++] = textureId;
            aIndex = adjustedAIndex * 4;
                    `;
                }
            }
            if (packer._inBatchIDAttrib) {
                packerBody += `
                    float32View[adjustedAIndex++] = factory._inBatchID;
                    aIndex = adjustedAIndex * 4;
                `;
            }
            if (packer._uniformIDAttrib) {
                packerBody += `
                    float32View[adjustedAIndex++] = factory._uniformID;
                    aIndex = adjustedAIndex * 4;
                `;
            }
        }
        else {
            if (!skipByteIndexConversion) {
                packerBody += `
                    adjustedAIndex = aIndex / 4;
                `;
            }
            packerBody += `
                    float32View[adjustedAIndex++] = factory._masterID;
                    aIndex = adjustedAIndex * 4;
            `;
        }
        packerBody += `}
            ${this.packer._indexProperty
            ? `const oldAIndex = this._aIndex;`
            : ''}
            this._aIndex = aIndex;
        `;
        if (this.packer._indexProperty) {
            packerBody += `
    const verticesBefore = oldAIndex / ${this.packer._vertexSize}
    const indexCount  = targetObject['${this.packer._indexProperty}'].length;

    for (let j = 0; j < indexCount; j++)
    {
        compositeIndices[iIndex++] = verticesBefore + targetObject['${this.packer._indexProperty}'][j];
    }

    this._iIndex = iIndex;
            `;
        }
        return new Function(...CompilerConstants.packerArguments, packerBody);
    }
    _compileSourceBufferExpression(redirect, i) {
        return (typeof redirect.source === 'string')
            ? `targetObject['${redirect.source}']`
            : `attributeRedirects[${i}].source(targetObject, factory._renderer)`;
    }
    _compileVertexCountExpression() {
        if (!this.packer._vertexCountProperty) {
            return `__buffer_0.length / ${this.packer._attribRedirects[0].size}`;
        }
        return ((typeof this.packer._vertexCountProperty === 'string')
            ? `targetObject.${this.packer._vertexCountProperty}`
            : `${this.packer._vertexCountProperty}`);
    }
    _sizeOf(i) {
        return PIXI.ViewableBuffer.sizeOf(this.packer._attribRedirects[i].type);
    }
};
export default BatchGeometryFactory;
//# sourceMappingURL=BatchGeometryFactory.js.map