import { StdBatchFactory } from './StdBatchFactory';
import { BatchGeometryFactory } from './BatchGeometryFactory';
import * as PIXI from 'pixi.js';
import { resolveConstantOrProperty, resolveFunctionOrProperty } from './resolve';
import { BatchDrawer } from './BatchDrawer';
export class BatchRenderer extends PIXI.ObjectRenderer {
    constructor(renderer, options) {
        super(renderer);
        this._attribRedirects = options.attribSet;
        this._indexProperty = options.indexProperty;
        this._vertexCountProperty = options.vertexCountProperty;
        this._textureProperty = options.textureProperty;
        this._texturesPerObject = typeof options.texturesPerObject !== 'undefined' ? options.texturesPerObject : 1;
        this._texIDAttrib = options.texIDAttrib;
        this._inBatchIDAttrib = options.inBatchIDAttrib;
        this._stateFunction = options.stateFunction || (() => PIXI.State.for2d());
        this._shaderFunction = options.shaderFunction;
        this._BatchFactoryClass = options.BatchFactoryClass || StdBatchFactory;
        this._BatchGeometryFactoryClass = options.BatchGeometryFactoryClass || BatchGeometryFactory;
        this._BatchDrawerClass = options.BatchDrawerClass || BatchDrawer;
        this._uniformRedirects = options.uniformSet || null;
        this._uniformIDAttrib = options.uniformIDAttrib;
        this._masterIDAttrib = options.masterIDAttrib;
        this.options = options;
        if (options.masterIDAttrib) {
            this._texIDAttrib = this._masterIDAttrib;
            this._uniformIDAttrib = this._masterIDAttrib;
            this._inBatchIDAttrib = this._masterIDAttrib;
        }
        this.renderer.runners.contextChange.add(this);
        if (this.renderer.gl) {
            this.contextChange();
        }
        this._objectBuffer = [];
        this._bufferedVertices = 0;
        this._bufferedIndices = 0;
        this._shader = null;
    }
    contextChange() {
        const gl = this.renderer.gl;
        if (PIXI.settings.PREFER_ENV === PIXI.ENV.WEBGL_LEGACY) {
            this.MAX_TEXTURES = 1;
        }
        else {
            this.MAX_TEXTURES = Math.min(gl.getParameter(gl.MAX_TEXTURE_IMAGE_UNITS), PIXI.settings.SPRITE_MAX_TEXTURES);
        }
        this._batchFactory = new this._BatchFactoryClass(this);
        this._geometryFactory = new this._BatchGeometryFactoryClass(this);
        this._drawer = new this._BatchDrawerClass(this);
    }
    start() {
        this._objectBuffer.length = 0;
        this._bufferedVertices = 0;
        this._bufferedIndices = 0;
        this._shader = this._shaderFunction(this);
    }
    render(displayObject) {
        this._objectBuffer.push(displayObject);
        this._bufferedVertices += this._vertexCountFor(displayObject);
        if (this._indexProperty) {
            this._bufferedIndices += resolveConstantOrProperty(displayObject, this._indexProperty).length;
        }
    }
    flush() {
        const { _batchFactory: batchFactory, _geometryFactory: geometryFactory, _stateFunction: stateFunction } = this;
        const buffer = this._objectBuffer;
        const bufferLength = buffer.length;
        batchFactory.reset();
        geometryFactory.init(this._bufferedVertices, this._bufferedIndices);
        let batchStart = 0;
        for (let objectIndex = 0; objectIndex < bufferLength;) {
            const target = buffer[objectIndex];
            const wasPut = batchFactory.put(target, resolveFunctionOrProperty(target, stateFunction));
            if (!wasPut) {
                batchFactory.build(batchStart);
                batchStart = objectIndex;
            }
            else {
                ++objectIndex;
            }
        }
        if (!batchFactory.ready()) {
            batchFactory.build(batchStart);
        }
        const batchList = batchFactory.access();
        const batchCount = batchFactory.size();
        let indices = 0;
        for (let i = 0; i < batchCount; i++) {
            const batch = batchList[i];
            const batchBuffer = batch.batchBuffer;
            const batchLength = batchBuffer.length;
            let vertexCount = 0;
            let indexCount = 0;
            batch.geometryOffset = indices;
            for (let j = 0; j < batchLength; j++) {
                const targetObject = batchBuffer[j];
                if (this._indexProperty) {
                    indexCount += resolveConstantOrProperty(targetObject, this._indexProperty).length;
                }
                else {
                    vertexCount += resolveConstantOrProperty(targetObject, this._vertexCountProperty);
                }
                geometryFactory.append(targetObject, batch);
            }
            batch.$vertexCount = vertexCount;
            batch.$indexCount = indexCount;
            indices += batch.$indexCount;
        }
        this._drawer.draw();
    }
    stop() {
        if (this._bufferedVertices) {
            this.flush();
        }
    }
    _vertexCountFor(targetObject) {
        return (this._vertexCountProperty)
            ? resolveConstantOrProperty(targetObject, this._vertexCountProperty)
            : resolveFunctionOrProperty(targetObject, this._attribRedirects[0].source).length
                / this._attribRedirects[0].size;
    }
}
export default BatchRenderer;
//# sourceMappingURL=BatchRenderer.js.map